#import "OggDecoder.h"
#define BUFFER_SIZE 4096

@implementation OggDecoder

- init
{
	playing = NO;
	stop = YES;
	return [super initWithDescription:@"OggVorbis Decoder"];
}

- (BOOL)isOurFile:(NSString *)filename
{
	NSString *ext;
	
	ext = [filename pathExtension];
	if (ext)
	{
		if ( [ext isEqualToString:@"ogg"] )
		{
			return YES;
		}
	}
	
	return NO;
}

- (NSDictionary *)getSongInfoForFile:(NSString *)filename
{
	OggVorbis_File ogg_to_get_comments;
	FILE *fp;
	int err;
	NSDictionary *result;
	
	// t@CJ
	fp = fopen([filename cString],"r");
	if (fp == NULL)
		return nil;
	
	err = ov_open(fp,&ogg_to_get_comments,NULL,0);
	if (err)
	{
		fclose(fp);
		return nil;
	}
	
	result = [self getSongInfoForOgg:&ogg_to_get_comments ofFile:filename];
	
	ov_clear(&ogg_to_get_comments);
	return result;
}

- (NSDictionary *)getSongInfoForOgg:(OggVorbis_File *)ogg_to_get_comments ofFile:(NSString *)filename
{
	NSCharacterSet *delimitor = [NSCharacterSet characterSetWithCharactersInString:@"="];
	vorbis_comment *comment;
	int i;
	NSMutableDictionary *result;
	NSString *title = nil,*artist = nil,*generated_title;
	
	comment = ov_comment(ogg_to_get_comments,-1);
	
	result = [NSMutableDictionary dictionary];
	for (i = 0;i < comment->comments;i++)
	{
		// KEY=VALUE
		// VALUEUTF-8B
		NSString *line = [NSString stringWithUTF8String:comment->user_comments[i]];
		NSRange pos_equal = [line rangeOfCharacterFromSet:delimitor];
		NSString *key,*value;
		
		if (pos_equal.length == 0) // NOT FOUND
			continue;
		
		key = [[line substringToIndex:pos_equal.location] uppercaseString]; // L[͑啶ɑĂ
		value = [line substringFromIndex:pos_equal.location+1];
		
		if ([key isEqualToString:@"TITLE"])
		{
			title = value;
		}
		else if ([key isEqualToString:@"ARTIST"])
		{
			artist = value;
			[result setObject:value forKey:@"artistName"];
		}
		else if ([key isEqualToString:@"ALBUM"])
		{
			[result setObject:value forKey:@"albumName"];
		}
	}
	
	if (artist != nil) // A[eBXgĂB
	{
		generated_title = [NSString stringWithFormat:@"%@ - %@",
							(title == nil ? [filename lastPathComponent] : title),
							artist];
	}
	else // A[eBXgs
	{
		generated_title = (title == nil ? [filename lastPathComponent] : title);
	}
	[result setObject:generated_title forKey:@"title"];
	[result setObject:generated_title forKey:@"songName"];
	[result setObject:[NSNumber numberWithInt:ov_time_total(ogg_to_get_comments,-1) * 1000] forKey:@"length"]; // Pʂ̓~bB
	
	return result;
}

- (BOOL)open_file:(NSString *)fileName
{
	FILE *fp;
	int err;
	
	fp = fopen([fileName cString],"r");
	if (fp == NULL)
		return NO;
	
	err = ov_open(fp,&ogg,NULL,0);
	if (err)
	{
		fclose(fp);
		return NO;
	}
	
	return YES;
}

- (void)decode_loop:(NSString *)filename
{
	vorbis_info *ogginfo;
	NSDictionary *info;
	char *buffer;
	NSAutoreleasePool *pool;
	pool = [[NSAutoreleasePool alloc] init];
	
	// t@CJ
	if ([self open_file:filename] == NO)
	{
		playing = NO;
		
		NSRunAlertPanel(@"%@",@"Couldn't open the ogg file.",@"OK",nil,getPackageName(), nil);
		[pool release];
		return;
	}
	ogginfo = ov_info(&ogg,-1);
	info = [self getSongInfoForOgg:&ogg ofFile:filename];

    if ( info == nil )
	{
		[pool release];
		return;
	}
	
	[Input setInfoTitle:[info objectForKey:@"title"]
				length:[[info objectForKey:@"length"] intValue] // Pʂ͂ǂ~b炵B
				rate:ogginfo->bitrate_nominal // C`LB{͕σrbg[gvZׂBd낤...
				frequency:ogginfo->rate
				numChannels:ogginfo->channels];
	
	// o̓foCXJ
	if(![[Output output] openAudioFormat:FMT_S16_NE
			       rate:ogginfo->rate
			       numChannels:ogginfo->channels])
	{
		playing = NO;
		ov_clear(&ogg);
	    NSRunAlertPanel(@"%@",@"Couldn't open audio output device.",@"OK",nil,getPackageName(), nil);
		
		[pool release];
	    return;
	}
	
	// obt@m
	buffer = malloc(BUFFER_SIZE);
	
	// [v
	while (stop == NO)
	{
		long read_bytes;
		int bitstream_id;
		
		if (seek_to != -1)
		{
			ov_time_seek(&ogg,(double)(seek_to / 1000.0));
			[[Output output] flush:seek_to];
			
			seek_to = -1;
		}
		
		read_bytes = ov_read(&ogg,buffer,BUFFER_SIZE,1,2,1,&bitstream_id);
		if (read_bytes == 0) // EOF
			break;
		
		read_bytes = [Effect modSampleData:(short int *)buffer
							length:read_bytes
							bitsPerSample:16
							numChannels:ogginfo->channels
							freq:ogginfo->rate];
		
		[Visualization addVisPcmTime:[[Output output] writtenTime]
							format:FMT_S16_NE
							numChannels:ogginfo->channels
							length:read_bytes
							data:buffer];
		
		[[Output output] writeAudioData:buffer length:read_bytes];
	}
	
	if (stop == NO) // ~w󂯂ɍŌ܂ōĐIB
		[[Output output] wait];
	
	ov_clear(&ogg);
	[[Output output] closeAudio];
	playing = NO;
	free(buffer);
	[self donePlaying:YES];
	
	[pool release];
}

- (void)playFile:(NSString *)filename
{
	stop = NO;
	playing = YES;
	[NSApplication detachDrawingThread:@selector(decode_loop:)
				toTarget:self
				withObject:[filename retain]];
}

- (void)stop
{
	stop = YES;
}

- (void)seek:(int)time
{
	// time : b
	seek_to = time * 1000;
	[[Output output] cancelWrite];
}

- (void)pause:(BOOL)p
{
	[[Output output] pause:p];
}

- (int)getTime
{
	return playing ? [[Output output] outputTime] : -1;
}

- (void)about
{
	
}

- (BOOL)hasAbout
{
	return NO;
}

- (void)configure
{
	
}

- (BOOL)hasConfigure
{
	return NO;
}

- (void)fileInfoBox:(NSString *)filename
{

}

- (BOOL)enabledByDefault
{
	return YES;
}

- (void)setEq:(BOOL)on preamp:(float)preamp bands:(float *)b
{
	
}

@end